package se.chalmers.tdv.ddmin;

import java.util.LinkedList;
import java.util.List;

public abstract class DD<G> {

	public static <G> List<G> minus(List<G> a, int splits, int slot) {
		List<G> result = new LinkedList<G>();
		int pos = 0;
		int length = a.size() / splits;

		while (pos < length * slot) {
			result.add(a.get(pos));
			pos++;
		}
		pos = pos + length;
		while (pos < a.size()) {
			result.add(a.get(pos));
			pos++;
		}
		return result;
	}

	/**
	 * Indicates that the test input failed
	 */
	public static final int FAIL = -1;
	/**
	 * Indicates that the test input passed
	 */
	public static final int PASS = +1;
	/**
	 * Placeholder for unimplemented tests
	 */
	public static final int UNRESOLVED = 0;

	/**
	 * Runs the program under test on the supplied input configuration
	 * 
	 * @param config
	 *            The input to test
	 * @return {@link DD#PASS} if the test was successful, {@link DD#FAIL}
	 *         otherwise. The method may return {@link DD#UNRESOLVED} for an
	 *         unimplemented test case, but this is treated the same as
	 *         {@link DD#PASS}.
	 */
	public abstract int test(List<G> config);

	/**
	 * Run the "ddmin" algorithm as published in Zeller, Andreas, and Ralf
	 * Hildebrandt. "Simplifying and isolating failure-inducing input." Software
	 * Engineering, IEEE Transactions on 28.2 (2002): 183-200.
	 * 
	 * @param _circumstances
	 *            Input circumstances for which the test failed
	 * @param debug
	 *            When true, prints debug messages to System.out
	 * @return Minimalised circumstances input for which the test fails.
	 */
	public List<G> ddmin(List<G> _circumstances, boolean debug) {

		List<G> circumstances = _circumstances;

		int n = 2;

		while (circumstances.size() >= 2) {

			boolean some_complement_is_failing = false;

			for (int i = n - 1; i >= 0; i--) {
				List<G> complement = DD.<G> minus(circumstances, n, i);

				if (debug)
					System.out.print("==> n: " + n + " i: " + i + " "
							+ complement);

				if (test(complement) == FAIL) {
					if (debug)
						System.out.println(" FAIL");
					circumstances = complement;
					n = Math.max(n - 1, 2);
					if (debug)
						System.out.println("Adjust number of chunks to " + n);
					some_complement_is_failing = true;
					break;
				}
				if (debug)
					System.out.println(" PASS");
			}

			if (!some_complement_is_failing) {
				if (n == circumstances.size()) {
					if (debug)
						System.out.println("Number of chunks is maximal!");
					break;
				}
				n = Math.min(n * 2, circumstances.size());
				if (debug)
					System.out.println("Increase number of chunks to " + n);
			}
		}

		return circumstances;
	}

}